using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.Unicode;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Microsoft.Win32.SafeHandles;
using Newtonsoft.Json;

namespace FitConnect.Encryption;

public class CertificateHelper {
    private readonly ILogger? _logger;

    public CertificateHelper(ILogger? logger = null) {
        _logger = logger;
    }

    internal bool ValidateCertificate(string keyJson, LogLevel logLevel,
        X509Certificate2[]? rootCertificate = null) =>
        ValidateCertificate(new JsonWebKey(keyJson), logLevel, rootCertificate);

    internal bool ValidateCertificate(X509Certificate2 certificate,
        out X509ChainStatus[] chainStatus,
        X509Certificate2[]? rootCertificate = null,
        LogLevel logLevel = LogLevel.Warning) {
        var certificateChain = new X509Chain();

//        certificate.ExportToPem($"./temp/{Guid.NewGuid().ToString()}");
        _logger?.LogDebug("Issuers: {Issuer}", certificate.Issuer);

        if (rootCertificate != null) {
            certificateChain.ChainPolicy.CustomTrustStore.AddRange(rootCertificate);
            certificateChain.ChainPolicy.TrustMode = X509ChainTrustMode.CustomRootTrust;
            certificateChain.ChainPolicy.RevocationMode = X509RevocationMode.Offline;
            certificateChain.ChainPolicy.RevocationFlag = X509RevocationFlag.EndCertificateOnly;
            _logger?.LogDebug("Using custom root certificate");
        }
        else {
            certificateChain.ChainPolicy.RevocationMode = X509RevocationMode.Online;
            certificateChain.ChainPolicy.RevocationFlag = X509RevocationFlag.EntireChain;
            certificateChain.ChainPolicy.VerificationFlags = X509VerificationFlags.NoFlag;
        }


        var result = certificateChain.Build(certificate);

        chainStatus = certificateChain.ChainStatus
            .Where(s => s.Status != X509ChainStatusFlags.PartialChain).ToArray();

        var statusAggregation = certificateChain.ChainStatus.Aggregate("",
            (r, s) => r + "\n\t - " + s.Status + ": " + s.StatusInformation);

        if (certificateChain.ChainStatus.Length > 0) {
            _logger?.Log(logLevel, "Certificate status: {ObjStatusInformation}",
                statusAggregation);
            // _logger?.Log(logLevel, "Chain status: {ChainStatus}", certificateChain.ToDisplayString());
        }

        return result;
    }

    internal bool ValidateCertificate(JsonWebKey key, LogLevel logLevel = LogLevel.Error,
        X509Certificate2[]? root = null) {
        var certificates = key.X5c.Select(s => new X509Certificate2(Convert.FromBase64String(s)))
            .ToList();

        // if (certificates.Count != 3) {
        //     _logger?.Log(logLevel, "Found {Count} certificate(s) but should be 3",
        //         certificates.Count);
        //     return false;
        // }
        // root ??= new X509Certificate2(Convert.FromBase64String(key.X5t));

        var valid = certificates.Aggregate(true,
            (result, cert) => result
                              && ValidateCertificate(cert, out _, root, logLevel)
            // && cert.Verify()
        );
        return valid;
    }
}

public static class X509Certificate2Extensions {
    public static void ExportToPem(this X509Certificate2 certificate, string fileName) {
        StringBuilder builder = new StringBuilder();
        builder.AppendLine("-----BEGIN CERTIFICATE-----");
        builder.AppendLine(
            Convert.ToBase64String(certificate.RawData, Base64FormattingOptions.InsertLineBreaks));
        builder.AppendLine("-----END CERTIFICATE-----");
        var content = builder.ToString();

        File.WriteAllText(fileName + ".pem", content);
        File.WriteAllText(fileName + ".json", JsonConvert.SerializeObject(certificate));
    }

    public static string ToDisplayString(this X509Chain chain) {
        var b = new StringBuilder();

        b.AppendLine($"[{nameof(chain.ChainPolicy.RevocationFlag)}]");
        b.AppendLine($"  {chain.ChainPolicy.RevocationFlag}");
        b.AppendLine();
        b.AppendLine($"[{nameof(chain.ChainPolicy.RevocationMode)}]");
        b.AppendLine($"  {chain.ChainPolicy.RevocationMode}");
        b.AppendLine();
        b.AppendLine($"[{nameof(chain.ChainPolicy.VerificationFlags)}]");
        b.AppendLine($"  {chain.ChainPolicy.VerificationFlags}");
        b.AppendLine();
        b.AppendLine($"[{nameof(chain.ChainPolicy.VerificationTime)}]");
        b.AppendLine($"  {chain.ChainPolicy.VerificationTime}");
        b.AppendLine();
        b.AppendLine($"[Application Policy]");
        foreach (var policy in chain.ChainPolicy.ApplicationPolicy) {
            b.AppendLine($"  {policy.ToDisplayString()}");
        }

        b.AppendLine();
        b.AppendLine($"[Certificate Policy]");
        foreach (var policy in chain.ChainPolicy.CertificatePolicy) {
            b.AppendLine($"  {policy.ToDisplayString()}");
        }

        b.AppendLine();
        b.AppendLine($"[Elements]");
        foreach (var (element, index) in chain.ChainElements.Cast<X509ChainElement>()
                     .Select((element, index) => (element, index))) {
            b.AppendLine();
            b.AppendLine($"[Element {index + 1}]");
            b.AppendLine();
            b.Append(element.Certificate.ToString());
            b.AppendLine();
            b.AppendLine($"[Status]");
            foreach (var status in element.ChainElementStatus) {
                b.AppendLine($"  {status.ToDisplayString()}");
            }
        }

        return b.ToString();
    }

    public static string ToDisplayString(this Oid oid) {
        return oid.FriendlyName == oid.Value
            ? $"{oid.Value}"
            : $"{oid.FriendlyName}: {oid.Value}";
    }

    public static string ToDisplayString(this X509ChainStatus status) {
        return $"{status.Status}: {status.StatusInformation}";
    }
}
